# © Copyright Databand.ai, an IBM Company 2022

import pandas as pd

from dbnd import task


@task
def task_that_spawns(value=1.0):
    # type: (float)-> float

    value_task = task_that_runs_inline.task(value=value)
    value_task.dbnd_run()

    return value_task.result.read_pickle() + 0.1


@task
def task_that_runs_inline(value=1.0):
    # type: (float)-> float
    return value + 0.1


@task
def func_return_df():
    return pd.DataFrame(data=[[3, 1]], columns=["c1", "c2"])


@task
def prepare_data(df):
    return df


@task
def func_pipeline(p: int):
    df = pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"])
    d1 = prepare_data(df)
    d2 = prepare_data(d1)
    return d2


@task
def func_pipeline2(p: int):
    df = func_return_df()
    d1 = func_gets(df)
    return d1


if __name__ == "__main__":
    import os

    os.environ["DBND__TRACKING"] = "true"
    func_pipeline2(4)



# © Copyright Databand.ai, an IBM Company 2022

from pathlib import Path

from dbnd_examples_orchestration.data import data_repo
from pandas import DataFrame
from sklearn.linear_model import ElasticNet

from dbnd import parameter, task
from targets.target_config import FileFormat


class TestDocInputsOutputs:
    def test_train_model(self):
        #### DOC START
        @task
        def train_model(
            training_set: DataFrame, alpha: float = 0.5, l1_ratio: float = 0.5
        ) -> ElasticNet:
            lr = ElasticNet(alpha=alpha, l1_ratio=l1_ratio)
            lr.fit(training_set.drop(["quality"], 1), training_set[["quality"]])
            return lr

        train_model.task(training_set=data_repo.wines).dbnd_run()
        #### DOC END

    def test_read_data(self):
        #### DOC START
        @task
        def read_data(path: Path) -> int:
            num_of_lines = len(open(path, "r").readlines())
            return num_of_lines

        #### DOC END
        read_data.task(path=data_repo.wines).dbnd_run()

    def test_prepare_data(self):
        #### DOC START
        @task
        def prepare_data(data: str) -> str:
            return data

        #### DOC END
        prepare_data.task(data="testing prepare dat").dbnd_run()

    def test_prepare_data_load_options(self):
        #### DOC START
        @task(data=parameter[DataFrame].csv.load_options(FileFormat.csv, sep="\t"))
        def prepare_data(data: DataFrame) -> DataFrame:
            data["new_column"] = 5
            return data

        #### DOC END
        prepare_data.dbnd_run(data_repo.wines)



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd_examples_orchestration.orchestration.tutorial_syntax.T01_tasks_and_pipelines import (
    calculate_values_pipeline,
    calculate_values_with_base_pipeline,
)


logger = logging.getLogger(__name__)


class TestTutorialSyntax(object):
    def test_pipe_operations_pipe(self):
        run_result = calculate_values_pipeline.dbnd_run()
        assert 2 == run_result.root_task.result[0].load(object)

    def test_pipe_operations_pipelines(self):
        run_result = calculate_values_with_base_pipeline.dbnd_run()
        assert 7 == run_result.root_task.result[2].load(object)



# © Copyright Databand.ai, an IBM Company 2022

"""#### DOC START
from dbnd import output, parameter
from dbnd_spark import SparkConfig, SparkTask


class WordCountTask(SparkTask):
    text = parameter.data
    counters = output

    main_class = WORD_COUNT_MAIN_CLASS
    # overides value of SparkConfig object
    defaults = {SparkConfig.driver_memory: "2.5g"}

    def application_args(self):
        return [self.text, self.counters]


#### DOC END"""



# © Copyright Databand.ai, an IBM Company 2022

# Inspired by an tensorflow tutorial:
#   https://www.tensorflow.org/guide/premade_estimators

import tensorflow as tf

from dbnd_examples_orchestration.orchestration.tool_tensorflow import iris
from sklearn.model_selection import train_test_split

from dbnd import ConfigPath, data, output, parameter
from dbnd_run.tasks import PipelineTask, PythonTask


CSV_COLUMN_NAMES = ["SepalLength", "SepalWidth", "PetalLength", "PetalWidth", "Species"]


class DownloadKeras(PythonTask):
    task_target_date = None
    url = parameter[str]
    downloaded = output.csv

    def run(self):
        tf.keras.utils.get_file(self.downloaded.path, self.url)


class PrepareTestAndValidationData(PythonTask):
    raw_data = data
    validation_size = parameter(default=0.5)

    test = output.csv
    validation = output.csv

    def run(self):
        raw = self.raw_data.read_df()
        test_df, validation_df = train_test_split(raw, test_size=self.validation_size)

        test_df.to_target(self.test, index=False)
        validation_df.to_target(self.validation, index=False)


class TrainIrisModel(PythonTask):
    train_set = data
    test_set = data

    batch_size = parameter[int]
    train_steps = parameter[int]

    model = output.folder
    model_data = output.folder

    def _load_data(self, y_name="Species"):
        train = self.train_set.read_df(names=CSV_COLUMN_NAMES, header=0)
        train_x, train_y = train, train.pop(y_name)

        test = self.test_set.read_df(names=CSV_COLUMN_NAMES, header=0)
        test_x, test_y = test, test.pop(y_name)

        return (train_x, train_y), (test_x, test_y)

    def run(self):
        (train_x, train_y), (test_x, test_y) = self._load_data()
        self.model_data.mkdir()
        classifier = iris.build_classifier(train_x, self.model_data.path)

        classifier.train(
            input_fn=lambda: iris.train_input_fn(train_x, train_y, self.batch_size),
            steps=self.train_steps,
        )

        evaluation = classifier.evaluate(
            input_fn=lambda: iris.eval_input_fn(test_x, test_y, self.batch_size)
        )

        for metric_name in evaluation.keys():
            self.log_metric(metric_name, evaluation[metric_name])

        self.model.mkdir_parent()
        t_model = classifier.export_savedmodel(
            self.get_target(name="tmp/model").path, iris.create_receiver_fn()
        )

        self.model.move_from(t_model)
        self.model_data.mark_success()

        # merged = tf.summary.merge_all()
        # train_writer = tf.summary.FileWriter(FLAGS.summaries_dir + '/train',
        #                                      sess.graph)
        # test_writer = tf.summary.FileWriter(FLAGS.summaries_dir + '/test')


class ValidateIrisModel(PythonTask):
    validation_set = data
    model = data

    report = output.csv

    def run(self):
        data = self.validation_set.read_df(names=CSV_COLUMN_NAMES, header=0)
        predictions = iris.predict(data=data, model_path=self.model.path)
        predictions.to_target(self.report)


class PredictIrisType(PipelineTask):
    batch_size = parameter.default(100)[int]
    train_steps = parameter.default(1000)[int]

    train_set_url = parameter(
        config_path=ConfigPath("dbnd_examples_orchestration", "iris_train_url")
    )[str]
    test_set_url = parameter(
        config_path=ConfigPath("dbnd_examples_orchestration", "iris_test_url")
    )[str]

    report = output

    def band(self):
        train = DownloadKeras(url=self.train_set_url)
        test = DownloadKeras(url=self.test_set_url)

        data = PrepareTestAndValidationData(raw_data=test.downloaded)

        model = TrainIrisModel(
            train_set=train.downloaded,
            test_set=data.test,
            batch_size=self.batch_size,
            train_steps=self.train_steps,
        )

        self.report = ValidateIrisModel(
            model=model.model, validation_set=data.validation
        ).report



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import logging
import random

from collections import defaultdict
from datetime import timedelta
from heapq import nlargest
from typing import List

import six

from dbnd import band, data, parameter, task
from dbnd.utils import data_combine, period_dates
from targets import target
from targets.types import DataList


@band(
    root_location=data.config("dbnd_examples_orchestration", "root_streams"),
    task_target_date=parameter[datetime.date],
)
def external_stream(root_location, task_target_date):
    return target(root_location, "%s.txt" % task_target_date.strftime("%Y-%m-%d"))


@task
def stream(seed=1):
    r = random.Random(x=seed)
    return [
        "{} {} {}".format(r.randint(0, 999), r.randint(0, 999), r.randint(0, 999))
        for _ in range(1000)
    ]


@task
def aggregate_artists(stream):
    # type: (DataList[str]) -> List[str]
    artist_count = defaultdict(int)
    for l in stream:
        _, artist, track = l.strip().split()
        artist_count[artist] += 1

    return [
        "{}\t{}".format(artist, count) for artist, count in six.iteritems(artist_count)
    ]


@task
def top_n_artists(artists, n_largest=10):
    # type: (DataList[str], int) -> List[str]
    def _input_iterator():
        for line in artists:
            artist, streams = line.strip().split()
            yield int(streams), int(artist)

    top_n = nlargest(n_largest, _input_iterator())
    return ["{}\t{}".format(streams, artist) for streams, artist in top_n]


@band
def top_artists_report(task_target_date, period=timedelta(days=2)):
    logging.info("top_artists_report")
    streams = [
        stream(task_name="Stream_%s" % i, task_target_date=d)
        for i, d in enumerate(period_dates(task_target_date, period))
    ]
    artists = aggregate_artists(stream=data_combine(streams))
    top_n = top_n_artists(artists=artists)
    return top_n


@band
def top_artists_big_report():
    return top_artists_report(period="3d", override={top_n_artists.task.n_largest: 15})



# © Copyright Databand.ai, an IBM Company 2022
import pytest

from dbnd._core.context.use_dbnd_run import disable_airflow_package


disable_airflow_package()


### MAIN FIXTURE, USED BY ALL TESTS ###
@pytest.fixture(autouse=True)
def dbnd_env_per_test(dbnd_run_pytest_env):
    yield dbnd_run_pytest_env



# © Copyright Databand.ai, an IBM Company 2022

"""
Code that goes along with the Airflow tutorial located at:
https://github.com/airbnb/airflow/blob/master/airflow/example_dags/tutorial.py
"""
from datetime import datetime, timedelta

from airflow import DAG
from airflow.operators.bash_operator import BashOperator

from dbnd_run.airflow.scheduler.dbnd_cmd_operators import dbnd_task_as_bash_operator
from dbnd_run.tasks.basics import dbnd_sanity_check


default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime.now(),
    "schedule_interval": timedelta(minutes=1),
    "email": ["airflow@example.com"],
    "email_on_failure": False,
    "email_on_retry": False,
    "retries": 1,
    "retry_delay": timedelta(minutes=5),
    # 'queue': 'bash_queue',
    # 'pool': 'backfill',
    # 'priority_weight': 10,
    # 'end_date': datetime(2016, 1, 1),
}

with DAG("scheduled_dbnd_check", default_args=default_args) as dag:
    templated_command = (
        "dbnd run dbnd_sanity_check --task-target-date {{ ds }} "
        "--name {{ params.my_param }}"
    )
    BashOperator(
        task_id="run_dbnd_check",
        bash_command=templated_command,
        params={"my_param": "name_from_params"},
        dag=dag,
    )

every_minute = dbnd_task_as_bash_operator(
    dbnd_sanity_check,
    name="dbnd_check_every_minute",
    schedule_interval=timedelta(minutes=2),
    default_args=default_args,
)

print(every_minute)



# © Copyright Databand.ai, an IBM Company 2022

from time import time

from dbnd_examples_orchestration.orchestration.tool_mxnet import (
    DownloadFile,
    MxNetGluonTask,
    fashion_data,
    read_minst,
)
from mxnet import autograd, gluon, init
from mxnet.gluon import nn
from mxnet.gluon.data.dataset import ArrayDataset
from mxnet.gluon.data.vision import transforms

from dbnd import data, namespace, output, parameter
from dbnd_run.tasks import PipelineTask


namespace("fashion")


def accuracy(output, label):
    return (output.argmax(axis=1) == label.astype("float32")).mean().asscalar()


def build_lenet_classifier():
    net = nn.Sequential()
    net.add(
        nn.Conv2D(channels=6, kernel_size=5, activation="relu"),
        nn.MaxPool2D(pool_size=2, strides=2),
        nn.Conv2D(channels=16, kernel_size=3, activation="relu"),
        nn.MaxPool2D(pool_size=2, strides=2),
        nn.Flatten(),
        nn.Dense(120, activation="relu"),
        nn.Dense(84, activation="relu"),
        nn.Dense(10),
    )
    net.initialize(init=init.Xavier())
    return net


def forward_backward_path(
    network, loss_function, acc_function, train_data, trainer, batch_size
):
    train_loss, train_acc = 0.0, 0.0
    for data, label in train_data:
        # forward + backward
        with autograd.record():
            output = network(data)
            loss = loss_function(output, label)
        loss.backward()
        # update parameters
        trainer.step(batch_size)
        # calculate training metrics
        train_loss += loss.mean().asscalar()
        train_acc += acc_function(output, label)
    return train_acc, train_loss


class TrainModel(MxNetGluonTask):
    learning_rate = parameter.value(0.1)
    optimizer = parameter.value("sgd")
    epoch_num = parameter.value(10)
    batch_size = 256

    training_data = data
    training_labels = data

    test_data = data
    test_labels = data

    model = output

    def _get_data_loader(self, data, labels):
        # prepare data
        data, label = read_minst(str(data), str(labels), gluon_format=True)
        transform_data = transforms.Compose(
            [transforms.ToTensor(), transforms.Normalize(0.13, 0.31)]
        )
        data_set = ArrayDataset(data, label).transform_first(transform_data)

        return gluon.data.DataLoader(data_set, batch_size=self.batch_size, shuffle=True)

    def get_trainer(self, network):
        return gluon.Trainer(
            network.collect_params(),
            self.optimizer,
            {"learning_rate": self.learning_rate},
        )

    def main(self, ctx):
        # prepare data
        train_dataloader = self._get_data_loader(
            self.training_data, self.training_labels
        )
        test_dataloader = self._get_data_loader(self.test_data, self.test_labels)

        network = build_lenet_classifier()
        trainer = self.get_trainer(network)

        # train
        for epoch in range(self.epoch_num):
            train_loss, train_acc, valid_acc = 0.0, 0.0, 0.0
            tic = time()
            train_acc, train_loss = forward_backward_path(
                network,
                gluon.loss.SoftmaxCrossEntropyLoss(),
                accuracy,
                train_dataloader,
                trainer,
                self.batch_size,
            )

            # calculate validation accuracy
            for data, label in test_dataloader:
                valid_acc += accuracy(network(data), label)

            self.log_metric(
                "epoch",
                (
                    epoch,
                    train_loss / len(train_dataloader),
                    train_acc / len(train_dataloader),
                    valid_acc / len(test_dataloader),
                    time() - tic,
                ),
            )

        network.save_parameters(self.model.path)


class PredictClothingType(PipelineTask):
    model = output

    def band(self):
        train_data = DownloadFile.from_web(fashion_data.train_data).data
        train_labels = DownloadFile.from_web(fashion_data.train_label).data
        test_data = DownloadFile.from_web(fashion_data.test_data).data
        test_labels = DownloadFile.from_web(fashion_data.test_label).data

        model_task = TrainModel(
            training_data=train_data,
            training_labels=train_labels,
            test_data=test_data,
            test_labels=test_labels,
        )
        self.model = model_task.model



# © Copyright Databand.ai, an IBM Company 2022

#### DOC START

"""[kubernetes_cluster_env]
_from = local
remote_engine = kubernetes_engine"""


"""[kubernetes_engine]
_type = kubernetes

container_repository = databand_examples
container_tag =

docker_build = True
docker_build_push = True

namespace = databand

debug = False
secrets = [
   { "type":"env", "target": "AIRFLOW__CORE__SQL_ALCHEMY_CONN", "secret" : "databand-secret", "key": "airflow_sql_alchemy_conn"},
   { "type":"env", "target": "AIRFLOW__CORE__FERNET_KEY", "secret" : "databand-secret", "key": "airflow_fernet_key"},
   { "type":"env", "target": "DBND__CORE__DATABAND_URL", "secret" : "databand-secret", "key": "databand_url"}
   ]

pod_exit_code_to_retry_count = { 255:3 }
pod_default_retry_delay = 3m"""

"""secrets = [   { "type":"env", "target": "AWS_ACCESS_KEY_ID", "secret" : "aws-secrets" , "key" :"aws_access_key_id"},
          { "type":"env", "target": "AWS_SECRET_ACCESS_KEY", "secret" : "aws-secrets" , "key" :"aws_secret_access_key"}]"""


#### DOC END



# © Copyright Databand.ai, an IBM Company 2022

import time

from dbnd_examples_orchestration.data import data_repo
from pandas import DataFrame
from sklearn.linear_model import ElasticNet

#### DOC START
from dbnd import task


@task
def train_model(
    training_set: DataFrame, alpha: float = 0.5, l1_ratio: float = 0.5
) -> ElasticNet:
    lr = ElasticNet(alpha=alpha, l1_ratio=l1_ratio)
    lr.fit(training_set.drop(["quality"], 1), training_set[["quality"]])
    return lr


def calculate_alpha():
    return "alpha=%s" % (time.time(),)


#### DOC END


class TestDocObjectConfiguration:
    def test_doc(self):
        train_model.dbnd_run(
            training_set=data_repo.wines, task_version=calculate_alpha()
        )



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import itertools
import logging

from typing import List, Tuple

import matplotlib
import numpy as np
import pandas as pd

from dbnd_examples_orchestration.data import demo_data_repo
from dbnd_examples_orchestration.orchestration.examples.wine_quality import (
    package_as_docker,
)
from matplotlib import figure
from sklearn import preprocessing
from sklearn.linear_model import ElasticNet
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from sklearn.model_selection import train_test_split

from dbnd import log_dataframe, log_metric, parameter, pipeline, task
from dbnd.utils import data_combine, period_dates
from targets import target


logger = logging.getLogger(__name__)
matplotlib.use("Agg")


def data_source(task_target_date, name):
    return target(demo_data_repo.partner_to_file(name, task_target_date))


@task
def get_and_enrich(raw_data: pd.DataFrame, column_name: str) -> pd.DataFrame:
    scaler = preprocessing.MinMaxScaler()
    raw_data[column_name + "_norm"] = scaler.fit_transform(
        raw_data[[column_name]].values.astype(float)
    )
    return raw_data


@task
def clean_data(raw_data: pd.DataFrame) -> pd.DataFrame:
    return raw_data.fillna(0)


@pipeline
def ingest_partner_a(task_target_date):
    raw_data = data_source(name="a", task_target_date=task_target_date)
    clean = clean_data(raw_data=raw_data)
    return get_and_enrich(raw_data=clean, column_name="0")


@pipeline
def ingest_partner_b(task_target_date):
    raw_data = data_source(name="b", task_target_date=task_target_date)
    data = clean_data(raw_data=raw_data)
    return data


@pipeline
def ingest_partner_c(task_target_date):
    raw_data = data_source(name="c", task_target_date=task_target_date)
    clean = clean_data(raw_data=raw_data)
    data = get_and_enrich(raw_data=clean, column_name="10")
    return data


@pipeline
def fetch_partner_data(
    task_target_date, selected_partners: List[str], period=datetime.timedelta(days=7)
) -> List[pd.DataFrame]:
    partner_data = []
    for partner in selected_partners:
        all_data = []
        for d in period_dates(task_target_date, period):
            if partner == "a":
                data = ingest_partner_a(task_target_date=d)
            elif partner == "b":
                data = ingest_partner_b(task_target_date=d)
            elif partner == "c":
                data = ingest_partner_c(task_target_date=d)
            else:
                raise Exception("Partner not found!")

            all_data.append(data)
        partner_data.append(data_combine(all_data, sort=True))
    return partner_data


@task
def calculate_features(
    raw_data: List[pd.DataFrame], selected_features: List[str] = None
) -> pd.DataFrame:
    result = raw_data.pop(0)
    for d in raw_data:
        result = result.merge(d, on="id")
    if selected_features:
        result = result[selected_features]
    return result


@task(result="training_set, test_set, validation_set")
def split_data(
    raw_data: pd.DataFrame,
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    columns_to_remove = set(["id", "0_norm", "10_norm"])
    if columns_to_remove.issubset(raw_data.columns):
        raw_data.drop(columns_to_remove, axis=1, inplace=True)

    train_df, test_df = train_test_split(raw_data)
    test_df, validation_df = train_test_split(test_df, test_size=0.5)

    log_dataframe("raw", raw_data)
    log_metric("target.mean", raw_data["target"].mean())
    log_metric("target.std", raw_data["target"].std())

    return train_df, test_df, validation_df


@task
def train_model(
    test_set: pd.DataFrame,
    training_set: pd.DataFrame,
    alpha: float = 1.0,
    l1_ratio: float = 0.5,
) -> ElasticNet:
    lr = ElasticNet(alpha=alpha, l1_ratio=l1_ratio)
    lr.fit(training_set.drop(["target"], 1), training_set[["target"]])
    prediction = lr.predict(test_set.drop(["target"], 1))

    (rmse, mae, r2) = calculate_metrics(test_set[["target"]], prediction)

    log_metric("rmse", rmse)
    log_metric("mae", mae)
    log_metric("r2", r2)

    logging.info(
        "Elasticnet model (alpha=%f, l1_ratio=%f): rmse = %f, mae = %f, r2 = %f",
        alpha,
        l1_ratio,
        rmse,
        mae,
        r2,
    )
    return lr


def _create_scatter_plot(actual, predicted) -> figure.Figure:
    import matplotlib.pyplot as plt

    fig = plt.figure()
    ax = fig.add_subplot(1, 1, 1)
    ax.set_title("Actual vs. Predicted")
    ax.set_xlabel("Actual Labels")
    ax.set_ylabel("Predicted Values")
    ax.scatter(actual, predicted)
    return fig


@task(result=("report", "prediction_scatter_plot"))
def validate_model_for_customer(
    model: ElasticNet, validation_dataset: pd.DataFrame, threshold=0.2
) -> Tuple[str, figure.Figure]:
    log_dataframe("validation", validation_dataset)
    # support for py3 parqeut
    validation_dataset = validation_dataset.rename(str, axis="columns")
    validation_x = validation_dataset.drop(["target"], 1)
    validation_y = validation_dataset[["target"]]

    prediction = model.predict(validation_x)
    (rmse, mae, r2) = calculate_metrics(validation_y, prediction)

    log_metric("rmse", rmse)
    log_metric("mae", mae)
    log_metric("r2", r2)
    fig = _create_scatter_plot(validation_y, prediction)
    if r2 < threshold:
        raise Exception(
            "Model quality is below threshold. Got R2 equal to %s, expect at least %s"
            % (r2, threshold)
        )

    return "%s,%s,%s" % (rmse, mae, r2), fig


@pipeline(result=("model", "validation"))
def train_model_for_customer(
    task_target_date,
    data: pd.DataFrame = None,
    alpha: float = 1.0,
    l1_ratio: float = 0.5,
    period=datetime.timedelta(days=7),
    selected_features: List[str] = None,
):
    if data is None:
        partners = fetch_partner_data(task_target_date=task_target_date, period=period)
        data = calculate_features(
            selected_features=selected_features, raw_data=partners
        )
    training_set, test_set, validation_set = split_data(raw_data=data)

    model = train_model(
        test_set=test_set, training_set=training_set, alpha=alpha, l1_ratio=l1_ratio
    )
    validation, _ = validate_model_for_customer(
        model=model, validation_dataset=validation_set
    )
    return model, validation


@pipeline
def train_for_all_customer(customer=parameter[List[pd.DataFrame]]):
    result = {}
    for c in customer:
        (model, validation) = train_model_for_customer(data=c)
        result[str(c)] = (model, validation)
    return result


@pipeline
def training_with_parameter_search(
    data: pd.DataFrame = demo_data_repo.sample_customer_file,
    alpha_step: float = 0.3,
    l1_ratio_step: float = 0.4,
):
    result = {}
    variants = list(
        itertools.product(np.arange(0, 1, alpha_step), np.arange(0, 1, l1_ratio_step))
    )
    logger.info("All Variants: %s", variants)
    for alpha_value, l1_ratio in variants:
        exp_name = "Predict_%f_l1_ratio_%f" % (alpha_value, l1_ratio)
        model, validation = train_model_for_customer(
            data=data, alpha=alpha_value, l1_ratio=l1_ratio, task_name=exp_name
        )

        result[exp_name] = (model, validation)
    return result


@pipeline(result=("model", "validation", "serving"))
def tran_model_and_package():
    model, validation = train_model_for_customer()
    serving = package_as_docker(model=model)
    return model, validation, serving


# Uttilities
def calculate_metrics(actual, pred):
    rmse = np.sqrt(mean_squared_error(actual, pred))
    mae = mean_absolute_error(actual, pred)
    r2 = r2_score(actual, pred)
    return rmse, mae, r2



# © Copyright Databand.ai, an IBM Company 2022

"""
class TestDocSparkOrchestration:
    def test_prepare_data_spark_task(self):
        #### DOC START
        import pyspark.sql as spark
        from dbnd import output, parameter
        from dbnd_spark.spark import spark_task

        @spark_task
        def prepare_data_spark(text) -> spark.DataFrame:
            from operator import add
            from dbnd_spark.spark import get_spark_session

            lines = text.rdd.map(lambda r: r[0])
            counts = (
                lines.flatMap(lambda x: x.split(" ")).map(lambda x: (x, 1)).reduceByKey(add)
            )

            return counts
        #### DOC END
        prepare_data_spark.dbnd_run(text = "Hello!")


    def test_prepare_data_pyspark_task(self):
        #### DOC START
        class PrepareDataPySparkTask(PySparkTask):
            text = parameter.data
            counters = output

            python_script = spark_folder("pyspark/word_count.py")

            def application_args(self):
                return [self.text, self.counters]
        #### DOC END
        PrepareDataPySparkTask(text=data_repo.wines.dbnd_run())

    def test_prepare_data_jvm_spark_task(self):
        #### DOC START
        class PrepareDataSparkTask(SparkTask):
            text = parameter.data
            counters = output

            main_class = "ai.databand.examples.WordCount"
            defaults = {
                SparkConfig.main_jar: "${DBND_HOME}/databand_examples/tool_spark/spark_jvm/target/ai.databand.examples-1.0-SNAPSHOT.jar"}

            def application_args(self):
                return [self.text, self.counters]
        #### DOC END

        PrepareDataSparkTask(text=data_repo.wines).dbnd_run()"""



# © Copyright Databand.ai, an IBM Company 2022

"""from dbnd import output, parameter
from dbnd_spark import spark, spark_task
from targets.target_config import FileFormat


#### DOC START
@spark_task(result=output.save_options(FileFormat.csv, header=True)[spark.DataFrame])
def custom_load_save_options(
    data=parameter.load_options(FileFormat.csv, header=False, sep="\t")[spark.DataFrame]
):
    print(data.show())
    return data

@spark_task
def prepare_data(data_path=parameter[PathStr]):
    df = get_spark_session()read.format("csv").options(header=False, sep="\t").load(data_path)
    print(data.show())
    return data
#### DOC END
"""



# © Copyright Databand.ai, an IBM Company 2022

import logging

from typing import Tuple

from dbnd import dbnd_tracking_start, pipeline, task


@task
def say_hello(text="sdfsd"):
    greeting = "Hey, {}!".format(text)
    logging.info(greeting)
    return greeting


@task
def join_greeting(base_greeting, extra_name):
    return "{} and {}".format(base_greeting, extra_name)


@pipeline
def say_hello_pipe(users_num=3):
    v = say_hello("some_user")
    for i in range(users_num):
        v = join_greeting(v, "user {}".format(i))

    return v


@task
def say_hello_to_everybody(users_num=3) -> Tuple[str, str]:
    v = ""
    for i in range(users_num):
        v = say_hello("user {}".format(i))

    hello_pipe = say_hello_pipe()
    return v, hello_pipe


if __name__ == "__main__":
    dbnd_tracking_start()
    say_hello_to_everybody()



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import logging

from time import sleep

from dbnd import config, databand_lib_path, override, pipeline, task
from dbnd_docker.docker.docker_task import DockerRunTask
from dbnd_run.tasks.basics import dbnd_sanity_check


logger = logging.getLogger(__name__)


@task(
    task_config=dict(
        kubernetes=dict(
            pod_yaml=override(
                databand_lib_path("conf", "kubernetes-pod-tensorflow.yaml")
            ),
            trap_exit_file_flag=override("/output/training_logs/main-terminated"),
            limits=override({"nvidia.com/gpu": 1}),
        )
    )
)
def task_with_custom_k8s_yml_gpu(
    check_time=datetime.datetime.now(), sleep_time_sec=120
):
    # type: ( datetime.datetime, int)-> str
    config.log_current_config(as_table=True)
    logger.info("Running Kube Sanity Check!")
    if sleep_time_sec:
        logger.info("sleeping for %s", sleep_time_sec)
        sleep(sleep_time_sec)
    return "Databand checked at %s" % check_time


class ExampleDockerNativeTask(DockerRunTask):
    command = "echo hi"
    image = "bash:4.4.23"


@task(
    task_config=dict(
        kubernetes=dict(
            # tolerations=[
            #     dict(
            #         key="special_gpu",
            #         operator="Equal",
            #         value="true",
            #         effect="NoSchedule",
            #     )
            # ],
        )
    )
)
def dbnd_kube_check(check_time=datetime.datetime.now(), sleep_time_sec=120):
    # type: ( datetime.datetime, int)-> str
    config.log_current_config(as_table=True)
    logger.info("Running Kube Sanity Check!")
    if sleep_time_sec:
        logger.info("sleeping for %s", sleep_time_sec)
        sleep(sleep_time_sec)
    return "Databand checked at %s" % check_time


@pipeline
def example_dockerized_pipeline():
    return {"native": ExampleDockerNativeTask(), "simple": dbnd_sanity_check()}


if __name__ == "__main__":
    t = dbnd_kube_check.task()
    t.dbnd_run()
    print(t)



# © Copyright Databand.ai, an IBM Company 2022

from __future__ import absolute_import

from dbnd_docker.docker.docker_task import DockerRunTask


class SimpleDockerizedPythonTask(DockerRunTask):
    command = "python -c 'import time;print(time.time())';time.sleep(100)"
    image = "python:3.9-slim"



# © Copyright Databand.ai, an IBM Company 2022

from dbnd_examples_orchestration.data import data_repo
from pandas import DataFrame

from dbnd import pipeline, task


@task
def prepare_data(raw_data: DataFrame = data_repo.wines):
    return raw_data


@task
def calculate_alpha(alpha: float = 0.5):
    return alpha


@task
def calculate_beta(beta: float = 0.1):
    return beta


@task
def calculate_coefficient(alpha, beta):
    return alpha / beta


class TestDocTasksWiringAtPipeline:
    def test_pipeline(self):
        ####DOC START
        @pipeline
        def evaluate_data(raw_data: DataFrame):
            data = prepare_data(raw_data=raw_data)
            alpha = calculate_alpha()
            beta = calculate_beta()
            coefficient = calculate_coefficient(alpha, beta)

            return data, coefficient

        @task
        def train_model(data: DataFrame, coefficient):
            model = data, coefficient
            return model

        @pipeline
        def create_model(raw_data: DataFrame):
            data, coefficient = evaluate_data(raw_data)
            return train_model(data, coefficient)

        #### DOC END

        create_model.dbnd_run(raw_data=data_repo.wines)

    def test_linear_reg_pipeline(self):
        @task
        def prepare_data() -> (str, str):
            return "training_set", "testing_set"

        @task
        def train_model(training_set):
            return "model"

        @task
        def test_model(model, testing_set):
            return "metrics"

        @task
        def get_validation_metrics(testing_set):
            return "validation_metrics"

        #### DOC START
        @pipeline(result=("model", "metrics", "validation_metrics"))
        def linear_reg_pipeline():
            training_set, testing_set = prepare_data()
            model = train_model(training_set)
            metrics = test_model(model, testing_set)
            validation_metrics = get_validation_metrics(testing_set)
            validation_metrics.task.set_upstream(metrics.task)
            # return the model along with metrics
            return model, metrics, validation_metrics

        #### DOC END

        linear_reg_pipeline.task().dbnd_run()



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import random

from collections import defaultdict
from datetime import timedelta
from heapq import nlargest

import six

from dbnd import data, output, parameter
from dbnd_run.tasks import DataSourceTask, PipelineTask, PythonTask
from targets import target


class ExternalStream(DataSourceTask):
    root_location = data.config("dbnd_examples_orchestration", "root_streams")
    task_target_date = parameter[datetime.date]

    stream = output

    def band(self):
        self.stream = target(
            self.root_location, "%s.txt" % self.task_target_date.strftime("%Y-%m-%d")
        )


class Stream(PythonTask):
    task_target_date = parameter[datetime.date]
    stream = output.data

    def run(self):
        lines = [
            "{} {} {}".format(
                random.randint(0, 999), random.randint(0, 999), random.randint(0, 999)
            )
            for _ in range(1000)
        ]
        self.stream.write("\n".join(lines))


class ArtistAggregator(PythonTask):
    streams = parameter.data
    index = output.data

    def run(self):
        artist_count = defaultdict(int)

        for s in self.streams:
            for l in s.readlines():
                _, artist, track = l.strip().split()
                artist_count[artist] += 1
        for artist, count in six.iteritems(artist_count):
            self.index.write("{}\t{}\n".format(artist, count))


class TopNArtists(PythonTask):
    artists = parameter.data
    n_largest = parameter.value(10)

    output = output.data

    def run(self):
        top_n = nlargest(self.n_largest, self._input_iterator())
        values = ["{}\t{}".format(streams, artist) for streams, artist in top_n]
        self.output.write("\n".join(values))

    def _input_iterator(self):
        for line in self.artists.readlines():
            artist, streams = line.strip().split()
            yield int(streams), int(artist)


class AggregateTopArtists(PipelineTask):
    period = parameter.value(timedelta(days=2))

    def band(self):
        streams = [
            Stream(task_name="Stream_%s" % i, task_target_date=d).stream
            for i, d in enumerate(period_dates(self.task_target_date, self.period))
        ]
        artists = ArtistAggregator(streams=streams)
        top_n = TopNArtists(artists=artists.index)
        return top_n



# © Copyright Databand.ai, an IBM Company 2022

# Python 3.6.8
import logging

from typing import Tuple

from pandas import DataFrame, Series
from sklearn import datasets
from sklearn.linear_model import LinearRegression
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import train_test_split

from dbnd import log_dataframe, log_metric, pipeline, task


logging.basicConfig(level=logging.INFO)


@task(result="training_data, testing_data")
def prepare_data() -> Tuple[DataFrame, DataFrame]:
    """load dataset from sklearn. split into training and testing sets"""
    raw_data = datasets.load_diabetes()

    # create a pandas DataFrame from sklearn dataset
    df = DataFrame(raw_data["data"], columns=raw_data["feature_names"])
    df["target"] = Series(raw_data["target"])

    # split the data into training and testing sets
    training_data, testing_data = train_test_split(df, test_size=0.25)

    # use DBND logging features to log DataFrames with histograms
    log_dataframe(
        "training data",
        training_data,
        with_histograms=True,
        with_schema=True,
        with_stats=True,
    )
    log_dataframe("testing_data", testing_data)

    return training_data, testing_data


@task(result="model")
def train_model(training_data: DataFrame) -> LinearRegression:
    """train a linear regression model"""
    model = LinearRegression()

    # train a linear regression model
    model.fit(training_data.drop("target", axis=1), training_data["target"])

    # use DBND log crucial details about the regression model with log_metric:
    log_metric("model intercept", model.intercept_)  # logging a numeric
    log_metric("coefficients", model.coef_)  # logging an np array
    return model


@task(result="performance_metrics")
def test_model(model: LinearRegression, testing_data: DataFrame) -> str:
    """test the model, output mean squared error and r2 score"""
    testing_x = testing_data.drop("target", axis=1)
    testing_y = testing_data["target"]
    predictions = model.predict(testing_x)
    mse = mean_squared_error(testing_y, predictions)
    r2_score = model.score(testing_x, testing_y)

    # use DBND log_metric to capture important model details:
    log_metric("mean squared error:", mse)
    log_metric("r2 score", r2_score)

    return f"MSE: {mse}, R2: {r2_score}"


@pipeline(result=("model", "metrics"))
def linear_reg_pipeline():
    training_set, testing_set = prepare_data()
    model = train_model(training_set)
    metrics = test_model(model, testing_set)

    # return the model along with metrics
    return model, metrics



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd_examples_orchestration.orchestration.tool_sklearn.sklearn_example import (
    linear_reg_pipeline,
)

from dbnd_run.testing.helpers import assert_run_task


logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def test_sklearn_tracking():
    task = assert_run_task(linear_reg_pipeline.task())
    assert task is not None
    assert task.result is not None



# © Copyright Databand.ai, an IBM Company 2022

import logging

from datetime import timedelta
from typing import List

from pandas import DataFrame

import dbnd

from dbnd import ParameterScope, data, output, parameter, task
from dbnd.utils import data_combine, period_dates
from dbnd_run.tasks import PythonTask
from targets.types import DataList


logger = logging.getLogger(__name__)

one_day = timedelta(days=1)

TUTORIAL = """

dbnd run FetchedData --task-env prod --period 10d

dbnd run ByDayExamplePipeline  --period "10d"  --FetchedData--disable-run

dbnd run ByDayExamplePipeline  --period "10d"  --FetchedData--disable-run


"""


def cb_data_dump_path(task, target_date, name):
    pass


class FetchIds(PythonTask):
    """Fetch ids by period"""

    period = parameter[timedelta]
    ids = output.prod_immutable[List[str]]
    labels = output.prod_immutable[List[str]]

    task_enabled_in_prod = False

    def run(self):
        values = ["a", "b", "c"]
        self.ids = ["%s-%s\n" % (s, self.task_target_date) for s in values]
        self.labels = ["%s-%s YES\n" % (s, self.task_target_date) for s in values]


class FetchData(PythonTask):
    """Fetch data of ids"""

    ids = data[List[str]]
    data = output.prod_immutable[List[str]]

    task_enabled_in_prod = False

    def run(self):
        values = ["a", "b", "c"]
        self.data = ["data-%s-%s" % (s, self.task_target_date) for s in values]


class IdsAndData(dbnd.PipelineTask):
    """Enrich Ids with data"""

    period = parameter.value(one_day)

    ids = output
    data = output

    def band(self):
        all_ids, all_data = {}, {}
        for i, d in enumerate(period_dates(self.task_target_date, self.period)):
            # if self.task_env == TaskEnv.prod and not self.run_on_prod:
            #     ids = cb_data_dump_path(task_target_date=d, name="ids")
            #     data = cb_data_dump_path(task_target_date=d, name="data")
            # else:
            ids = FetchIds(task_target_date=d, period=one_day).ids
            data = FetchData(task_target_date=d, ids=ids).data

            d_key = d.strftime("%Y-%m-%d")
            all_ids[d_key] = ids
            all_data[d_key] = data
        self.ids = data_combine(all_ids.values(), sort=True)
        self.data = data_combine(all_data.values(), sort=True)


class ProductionIdsAndData(IdsAndData):
    """Enrich Ids with data (Production)"""

    task_env = "aws_prod"


class Features(PythonTask):
    ids = data.target
    data = data.target

    features = output

    def run(self):
        features = []

        for ids, data in zip(self.ids.list_partitions(), self.data.list_partitions()):
            logger.info("Working on %s, %s", ids, data)
            # we put features one by one with ids
            partition = [
                "%s -> %s" % (k, v)
                for k, v in zip(
                    ids.load(value_type=List[str]), data.load(value_type=List)
                )
            ]

            features.extend(partition)
        self.features = features


def features_df(ids, data):
    # type: (DataFrame, DataFrame) -> DataFrame
    pass


@task
def features_simple(ids, data):
    # type: (DataList[str],DataList[str]) -> List[str]
    """Create features"""
    return ["%s -> %s\n" % (k.strip(), v.strip()) for k, v in zip(ids, data)]


class FeaturesPartitioned(PythonTask):
    """Create partitioned features"""

    ids = parameter.data
    data = parameter.data
    features = output

    def run(self):
        features = []

        for ids, data in zip(self.ids.list_partitions(), self.data.list_partitions()):
            logger.info("Working on %s, %s", ids, data)
            # we put features one by one with ids
            partition = [
                "%s -> %s\n" % (k, v)
                for k, v in zip(ids.load(List[str]), data.load(List[str]))
            ]

            features.extend(partition)
        self.features = features


class ByDayExamplePipeline(dbnd.PipelineTask):
    """Entry point of partitioned_by_day"""

    period = parameter(default=one_day, scope=ParameterScope.children)[timedelta]
    features = output

    partitioned_flow = parameter(
        default=False, description="Run features in paritioned mode"
    )[bool]

    def band(self):
        fetched = ProductionIdsAndData(period=self.period)

        if self.partitioned_flow:
            features_task = FeaturesPartitioned(ids=fetched.ids, data=fetched.data)
        else:
            features_task = features_simple(ids=fetched.ids, data=fetched.data)

        self.features = features_task



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import logging
import sys

from random import shuffle


logger = logging.getLogger(__name__)


def run_spark(args):
    from pyspark.sql import SparkSession
    from pyspark.sql.types import StringType, StructField, StructType

    timestamp = str(datetime.datetime.now()).replace(" ", "_")

    spark = SparkSession.builder.appName("Task_%s" % timestamp).getOrCreate()
    sc = spark.sparkContext

    chopped = []

    vegg = sc.textFile(args[1]).collect()
    logger.info("Got {}. Start Chopping.".format(",".join(vegg)).replace("\n", ""))

    for line in vegg:
        chopped.extend(list(line.rstrip()))

    shuffle(chopped)

    result = "".join(chopped)
    logger.info("Chopped vegetables:" + result)

    fields = [StructField("salad", StringType(), True)]
    schema = StructType(fields)

    df = spark.createDataFrame([(result,)], schema)
    df.coalesce(1).write.csv(args[2])
    sc.stop()


if __name__ == "__main__":
    run_spark(sys.argv)



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd import PythonTask, output, parameter


logger = logging.getLogger()


class GenerateReportToCustomLocation(PythonTask):
    _conf__base_output_path_fmt = (
        "{root}/{env_label}/reports/{name}/"
        "{output_name}{output_ext}/date={task_target_date}"
    )
    name = parameter.value("report")

    report = output[str]

    def run(self):
        logger.info("Going to write to %s", self.report)
        self.report = "Some text in weird location!"



# © Copyright Databand.ai, an IBM Company 2022

from typing import NamedTuple

import pandas as pd

from pandas import DataFrame

from dbnd import task


@task(result=("features", "scores"))
def f_returns_two_dataframes_v1(p: int) -> (DataFrame, DataFrame):
    return (
        pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
        pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
    )


@task(result="features,scores")
def f_returns_two_dataframes_v2(p: int) -> (DataFrame, DataFrame):
    return (
        pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
        pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
    )


@task
def f_returns_two_dataframes_no_hint(p: int) -> (DataFrame, DataFrame):
    return (
        pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
        pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
    )


FeatureStore = NamedTuple("FeatureStore", features=DataFrame, scores=DataFrame)


@task
def f_returns_two_dataframes_named_tuple_v1(p: int) -> FeatureStore:
    return FeatureStore(
        pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
        pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
    )


@task
def f_returns_two_dataframes_named_tuple_v2(
    p: int,
) -> NamedTuple("FeatureStore", features=DataFrame, scores=DataFrame):
    return (
        pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
        pd.DataFrame(data=[[p, 1]], columns=["c1", "c2"]),
    )


@task
def f_multiple_outputs_py2():
    # type: ()-> (pd.DataFrame, pd.DataFrame)
    return (
        pd.DataFrame(data=[[1, 1]], columns=["c1", "c2"]),
        pd.DataFrame(data=[[1, 1]], columns=["c1", "c2"]),
    )


@task
def f_returns_huge_dataframe(p: int) -> DataFrame:
    return pd.DataFrame(data=[[p] * 100000])



# © Copyright Databand.ai, an IBM Company 2022

import fnmatch
import logging
import re

from airflow.hooks.S3_hook import S3Hook
from airflow.operators.sensors import BaseSensorOperator

from targets.utils.path import path_to_bucket_and_key


logger = logging.getLogger(__name__)

"""
This file contains Airflow operator and hook to monitor S3 path for new file.
Used by scheduled_s3_sensor DAG in  airflow/scheduled_sensor.py
"""


class S3SyncHook(S3Hook):
    def list_wildcard_keys(self, wildcard_key, bucket_name=None, delimiter=""):
        if not bucket_name:
            (bucket_name, wildcard_key) = self.parse_s3_url(wildcard_key)

        prefix = re.split(r"[*]", wildcard_key, 1)[0]
        klist = self.list_keys(bucket_name, prefix=prefix, delimiter=delimiter)
        if klist:
            return [k for k in klist if fnmatch.fnmatch(k, wildcard_key)]
        return []

    def diff(self, source, destination):
        source_bucket, source_key = path_to_bucket_and_key(source)
        destination_bucket, destination_prefix = path_to_bucket_and_key(destination)
        destination_prefix += "/"

        source_state = self.list_wildcard_keys(source_key, source_bucket)
        destination_state = (
            self.list_keys(bucket_name=destination_bucket, prefix=destination_prefix)
            or []
        )
        logger.info("Full destination state: %s", destination_state)
        if destination_prefix:
            destination_state = [
                d[len(destination_prefix) :] for d in destination_state
            ]

        logging.info("source list:  {source_state}".format(**locals()))
        logging.info("destination list: {destination_state}".format(**locals()))

        diff = list(set(source_state) - set(destination_state))

        return ["s3://" + source_bucket + "/" + key for key in diff]


class S3StatefulSensor(BaseSensorOperator):
    from airflow.utils.decorators import apply_defaults

    template_fields = ("source", "destination")

    @apply_defaults
    def __init__(
        self, destination, source, aws_conn_id="aws_default", verify=None, **kwargs
    ):
        self.source = source
        self.destination = destination

        self.aws_conn_id = aws_conn_id
        self.verify = verify

        super(S3StatefulSensor, self).__init__(**kwargs)

    def poke(self, context):
        hook = S3SyncHook(aws_conn_id=self.aws_conn_id, verify=self.verify)

        logging.info(
            "Poking for key : {self.source}, syncing with: {self.destination})".format(
                **locals()
            )
        )

        diff = hook.diff(self.source, self.destination)
        if diff:
            return True
        return False



# © Copyright Databand.ai, an IBM Company 2022

import pytest
import six

from dbnd_examples_orchestration.data import data_repo, dbnd_examples_data_path
from dbnd_examples_orchestration.orchestration.examples import wine_quality
from dbnd_examples_orchestration.orchestration.examples.wine_quality_as_task_class import (
    PredictWineQuality,
    PredictWineQualityParameterSearch,
)

from dbnd import dbnd_config, relative_path
from dbnd.testing.helpers import run_test_notebook
from dbnd_run.testing.helpers import assert_run_task


class TestWineQualityClasses(object):
    def test_wine_quality_class(self):
        task = PredictWineQualityParameterSearch(alpha_step=0.4)
        assert_run_task(task)

    def test_wine_quality_deco_search(self):
        task = wine_quality.predict_wine_quality_parameter_search.t(
            alpha_step=0.5,
            override={wine_quality.predict_wine_quality.t.data: data_repo.wines},
        )
        assert_run_task(task)

    def test_wine_quality_deco_with_custom_data(self):
        task = wine_quality.predict_wine_quality.t(alpha=0.5, data=data_repo.wines)
        assert_run_task(task)

    @pytest.mark.skipif(not six.PY3, reason="requires python3")
    def test_wine_quality_deco_simple_all(self):
        with dbnd_config(
            {"local_prod": {"_from": "local", "env_label": "prod", "production": True}}
        ):
            task = wine_quality.predict_wine_quality.t(
                alpha=0.5, override={wine_quality.fetch_data.t.task_env: "local_prod"}
            )
            assert_run_task(task)

    def test_wine_quality_deco_simple(self):
        task = wine_quality.predict_wine_quality.t(alpha=0.5)
        assert_run_task(task)

    def test_wine_quality_gz(self):
        task = PredictWineQuality(data=dbnd_examples_data_path("wine_quality.csv.gz"))
        assert_run_task(task)

    def test_prepare_data(self):
        task = wine_quality.prepare_data.t(
            raw_data=dbnd_examples_data_path("wine_quality.csv.gz")
        )
        assert_run_task(task)

    # TODO: Fix this, it works locally in tox but not in CI
    # https://app.asana.com/0/1199880094608584/1200788284410456
    @pytest.mark.skip("doesnt pass in ci")
    def test_notebook(self):
        ipynb = relative_path(wine_quality.__file__, "wine_quality_as_notebook.ipynb")
        run_test_notebook(ipynb)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd_examples_orchestration.data import data_repo

from dbnd import PipelineTask, Task, output, parameter


def do_something_with_data(data):
    return data


class TestDocClassDefinitionsAsAClass:
    def test_doc(self):
        #### DOC START
        class PrepareData(Task):
            data = parameter.data

            prepared_data = output.csv.data

            def run(self):
                self.prepared_data = do_something_with_data(self.data)

        class PrepareDataPipeline(PipelineTask):
            data = parameter.data

            prepared_data = output.csv.data

            def band(self):
                self.prepared_data = PrepareData(data=self.data).prepared_data

        #### DOC START

        PrepareDataPipeline(data=data_repo.wines).band()

    def test_task_decorator(self):
        #### DOC START
        from dbnd import task

        @task
        class PrepareData:
            data = parameter.data

            prepared_data = output.csv.data

            def run(self):
                self.prepared_data = do_something_with_data(self.data)

        #### DOC START



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd import PipelineTask, PythonTask, output, parameter
from targets.dir_target import DirTarget
from targets.target_config import file
from targets.values import TargetValueType


logger = logging.getLogger(__name__)


class MyCustomFolder(DirTarget):
    def __init__(self, *args, **kwargs):
        super(MyCustomFolder, self).__init__(*args, **kwargs)
        self.scores = self.partition("scores", config=file.csv)
        self.reports = self.partition("reports", config=file.csv)


my_custom_target = parameter.type(TargetValueType).custom_target(
    MyCustomFolder, folder=True
)


class TaskWithCustomOutput(PythonTask):
    custom = my_custom_target.output

    def run(self):
        self.custom.scores.write("1")
        self.custom.reports.write("2")
        self.custom.mark_success()


class TaskWithCustomInput(PythonTask):
    custom = my_custom_target
    t_output = output.data

    def run(self):
        assert self.custom.scores.read() == "1"
        assert self.custom.reports.read() == "2"
        self.t_output.write("done")


class CustomIOPipeline(PipelineTask):
    t_output = output.data

    def band(self):
        custom = TaskWithCustomOutput().custom
        self.t_output = TaskWithCustomInput(custom=custom).t_output



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import logging

from random import shuffle
from typing import List

from dbnd_examples_orchestration.data import data_repo

from dbnd import log_metric, output, pipeline, task
from dbnd.errors import TaskValidationError
from targets import Target
from targets.types import DataList


#
# language=bash
TUTORIAL = """
# Execute without databand
python dbnd_examplesExecute full pipeline
dbnd run prepare_salad

# Execute task
dbnd run cut --vegetables data/vegetables_for_pasta_salad.txt
"""


@task
def add_dressing(chopped_vegetables, dressing, salt_amount="low"):
    # type: (DataList[str], str, str) -> List[str]
    # adds dressing
    if salt_amount == "very_high":
        raise TaskValidationError("Salt level is too high!")

    mix = [
        "%s->%s  (%s salt)" % (dressing, v.rstrip(), salt_amount)
        for v in chopped_vegetables
    ]

    log_metric("dressed", len(mix))
    log_metric("dressing", dressing)
    log_metric("dict", {"a": datetime.datetime.utcnow(), "tuple": ("a", 1)})

    logging.info("Dressing result %s", mix)
    return [x + "\n" for x in mix]


@task(result=output().data_list_str)
def cut(vegetables):
    # type: (DataList[str]) -> DataList[str]
    # cuts vegetables
    chopped = []

    logging.info(
        "Got {}. Start Chopping.".format(",".join(vegetables)).replace("\n", "")
    )

    for line in vegetables:
        chopped.extend(list(line.rstrip()))

    shuffle(chopped)

    logging.info("Chopped vegetables:" + " ".join(chopped))

    return [x + "\n" for x in chopped]


@pipeline
def prepare_salad(vegetables=data_repo.vegetables, dressing="oil"):
    # type: (Target, str) -> List[str]
    return add_dressing(cut(vegetables), dressing)


if __name__ == "__main__":
    with open(data_repo.vegetables, "r") as f:
        content = f.readlines()

    salad = prepare_salad(content, "mayo")

    with open("output.txt", "w") as f:
        f.writelines(salad)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd import task


class TestDocConfigurationLayers:
    def test_doc(self):
        #### DOC START
        @task
        def calculate_alpha(alpha=0.5, beta=0.1):
            return alpha

        alpha_task = calculate_alpha(alpha=4)
        #### DOC END



# © Copyright Databand.ai, an IBM Company 2022

import datetime

from datetime import timedelta

from dbnd_examples_orchestration.data import data_repo

import dbnd

from dbnd import data, output, parameter
from dbnd.utils import data_combine, period_dates
from dbnd_run.tasks import DataSourceTask, PythonTask
from targets import target
from targets.target_config import folder


class RawDeviceLog(DataSourceTask):
    """raw device logs for partitioned pipeline"""

    task_target_date = parameter[datetime.date]
    root_location = data.default(data_repo.partitioned_data)

    logs = output

    def band(self):
        self.logs = target(
            self.root_location,
            "%s/" % self.task_target_date.strftime("%Y-%m-%d"),
            config=folder.without_flag(),
        )


class DeviceLogProjection(PythonTask):
    """normalize device logs for partitioned pipeline"""

    raw_logs = data.target

    projected_logs = output.target

    def run(self):
        self.projected_logs.write("\n".join(self.raw_logs.readlines()))


class DeviceLogsPipeline(dbnd.PipelineTask):
    """project device logs for partitioned pipeline"""

    period = parameter(default=timedelta(days=2))[timedelta]
    projected = output

    def band(self):
        projected_logs = []
        for i, d in enumerate(period_dates(self.task_target_date, self.period)):
            raw_logs = RawDeviceLog(task_target_date=d).logs
            projected = DeviceLogProjection(raw_logs=raw_logs, task_target_date=d)
            projected_logs.append(projected.projected_logs)
        self.projected = data_combine(projected_logs)


class Features(PythonTask):
    """create features for partitioned pipeline"""

    projected_logs = data.target
    features = output.target

    def run(self):
        data = self.projected_logs.readlines()
        self.features.write("".join(data))


class ExamplePartitionedPipeline(dbnd.PipelineTask):
    """Entry point for partitioned pipeline"""

    features = output.target

    def band(self):
        projected = DeviceLogsPipeline().projected
        self.features = Features(projected_logs=projected).features



# © Copyright Databand.ai, an IBM Company 2022

from datetime import datetime, timedelta

from airflow import DAG
from dbnd_examples_orchestration.orchestration.dbnd_aws import (
    partner_data_ingest_new_files,
)
from dbnd_examples_orchestration.orchestration.dbnd_aws.sync_operators import (
    S3StatefulSensor,
)

from dbnd_run.airflow.scheduler.dbnd_cmd_operators import dbnd_task_as_bash_operator


"""
    this file should be put to airflow dag folder (or dag folder should point here)
    It contains definition of Airflow Dag that is used to periodically schedule databand runs.
    A scheduled_s3_sensor DAG is running Airflow sensor that waits and fires event when new files
    are added to source path (new means it does not exists in destination path). ingest_op is running as downstream
    task and ingest new files after validation into destination path. This DAG is scheduled to run every 2 minutes
    note that ingest_op is defines as databand task and used as Airflow operator.
    """

default_args = {
    "owner": "airflow",
    "depends_on_past": False,
    "start_date": datetime(year=2019, month=5, day=23, hour=13),
    "email": ["airflow@example.com"],
    "email_on_failure": False,
    "email_on_retry": False,
    "retries": 1,
    "retry_delay": timedelta(minutes=5),
    "params": {
        "source": "s3://databand-victor-test/sensor2/*.csv",
        "destination": "s3://victor-test-target/tmp/3",
    }
    # 'queue': 'bash_queue',
    # 'pool': 'backfill',
    # 'priority_weight': 10,
    # 'end_date': datetime(2016, 1, 1),
}

with DAG(
    "scheduled_s3_sensor",
    schedule_interval=timedelta(minutes=2),
    max_active_runs=1,
    default_args=default_args,
) as dag:

    sensor = S3StatefulSensor(
        task_id="trigger_on_new_files",
        source="{{params.source}}",
        destination="{{params.destination}}",
        timeout=18 * 60 * 60,
        poke_interval=120,
        dag=dag,
    )

    ingest_op = dbnd_task_as_bash_operator(
        partner_data_ingest_new_files,
        cmd_line="--task-target-date {{ ds }} "
        "--source '{{params.source}}' "
        "--destination '{{params.destination}}' ",
    )

    ingest_op.set_upstream(sensor)



# © Copyright Databand.ai, an IBM Company 2022

import logging

from pandas import DataFrame

from dbnd import output, parameter, task
from targets import DataTarget, Target
from targets.types import Path


logger = logging.getLogger(__name__)


@task(some_output=output.folder)
def plot_graphs(data, some_output):
    # type: (DataFrame, DataTarget) -> DataFrame
    some_output.partition().dump(data)
    some_output.mark_success()
    return data.tail(1)


@task(some_output=output.folder)
def plot_graphs_only(data, some_output):
    # type: (DataFrame, DataTarget) -> None
    some_output.partition().dump(data)


@task(folder_input=parameter.folder[Target])
def read_folder_status(folder_input):
    # type: (Target) -> str
    return str(folder_input.list_partitions())


@task(folder_input=parameter.folder.with_flag("MY_SUCCESS")[Path])
def read_folder_status_custom_flag(folder_input):
    # type: (Path) -> str
    v = str(list(folder_input.glob("*")))
    print(v)
    return v


@task(folder_input=parameter.folder.with_flag(False)[str])
def read_folder_status_noflag(folder_input):
    # type: (str) -> str
    v = str(list(Path(folder_input).glob("*")))
    print(v)
    return v



# © Copyright Databand.ai, an IBM Company 2022

import datetime

from pathlib import Path
from typing import List

from dbnd_examples_orchestration.data import data_repo
from pandas import DataFrame

from dbnd import parameter, task
from targets import Target
from targets.types import PathStr


class TestDocTaskParameters:
    def test_prepare_data_default(self):
        #### DOC START
        @task
        def prepare_data(data: DataFrame = None) -> DataFrame:
            return data

        #### DOC END
        prepare_data.dbnd_run(data=data_repo.wines)

    def test_prepare_data_parameter_factory(self):
        #### DOC START
        @task
        def prepare_data(data=parameter.default(None)[DataFrame]) -> DataFrame:
            return data

        #### DOC END
        prepare_data.dbnd_run(data=data_repo.wines)

    def test_prepare_data_parameter_decorator(self):
        #### DOC START
        @task(data=parameter.default(None)[DataFrame])
        def prepare_data(data) -> DataFrame:
            return data

        #### DOC END
        prepare_data.dbnd_run(data=data_repo.wines)

    def test_calculate_alpha_value_factory(self):
        #### DOC START
        @task(alpha=parameter.value(0.5))
        def calculate_alpha(alpha) -> float:
            return alpha

        #### DOC END
        calculate_alpha.dbnd_run()

    def test_prepare_data_simple_types(self):
        #### DOC START
        @task
        def prepare_data(v_int: int, v_bool: bool, v_str: str) -> str:
            pass

        #### DOC END
        prepare_data.dbnd_run(v_int=1, v_bool=True, v_str="hello!")

    def test_prepare_data_datetime_types(self):
        #### DOC START
        @task
        def prepare_data(
            v_datetime: datetime.datetime,
            v_date: datetime.date,
            v_period: datetime.timedelta,
        ) -> str:
            pass

        #### DOC END
        prepare_data.dbnd_run(
            v_datetime=datetime.datetime.now(),
            v_date=datetime.date.today(),
            v_period=datetime.timedelta(days=51),
        )

    def test_prepare_data_path_types(self):
        #### DOC START
        @task
        def prepare_data(
            pathlib_path: Path, str_as_path: PathStr, target_path: Target
        ) -> str:
            assert isinstance(str_as_path, str)

        #### DOC END

    def test_prepare_data_various_lists(self):
        #### DOC START
        @task
        def prepare_data(
            v_list: List,
            v_list_obj: List[object],
            v_list_str: List[str],
            v_list_int: List[int],
            v_list_date: List[datetime.date],
        ) -> str:
            assert isinstance(v_list_date[0], datetime.date)

            assert isinstance(v_list_obj[0], int)
            assert isinstance(v_list_obj[1], str)

            assert isinstance(v_list_int[0], int)
            assert isinstance(v_list_str[0], str)

            return "OK"

        #### DOC END

    def test_insignificant_parameters(self):
        #### DOC START
        from dbnd import parameter

        @task
        def calculate_alpha(
            alpha=parameter[float],
            debug_level=parameter(significant=False).value(0)[int],
        ):
            return alpha

        #### DOC END
        calculate_alpha.dbnd_run(alpha=0.5)



# © Copyright Databand.ai, an IBM Company 2022

from typing import List

from pyspark.sql import DataFrame

from dbnd import data_source_pipeline, extend, parameter, pipeline
from dbnd_run.task.data_source_task import data_combine
from dbnd_spark import SparkConfig, spark_task
from targets import target
from targets.multi_target import MultiTarget
from targets.target_config import FileFormat, TargetConfig


# Spark doesn't have "tsv", format, so we fake it, in case parameter will get csv we will read it as tsv
tsv_df = parameter.tsv.load_options(FileFormat.csv, delimiter="\t").save_options(
    FileFormat.csv, delimiter="\t", header=True
)[DataFrame]
parquet_df = parameter.parquet[DataFrame]


@data_source_pipeline
def existing_data(
    root_path: str,
    config_ids_A: List[int],
    config_ids_B: List[int],
    config_ids_C: List[int],
) -> (MultiTarget, MultiTarget, MultiTarget):
    base_target = target(root_path)

    def get_multi_target_from_config_ids(config_ids):
        configs = []
        # multi-target config is based on the first target in the list
        # we use csv format, as spark doesn't have explicit tsv support
        target_config = TargetConfig(folder=True, format=FileFormat.csv, flag=None)
        for config_id in config_ids:
            config = target(base_target, f"configID={config_id}", config=target_config)
            configs.append(config)
        return data_combine(configs)

    data_A = get_multi_target_from_config_ids(config_ids_A)
    data_B = get_multi_target_from_config_ids(config_ids_B)
    data_C = get_multi_target_from_config_ids(config_ids_C)
    return data_A, data_B, data_C


@spark_task(
    data_a=tsv_df,
    data_b=tsv_df,
    data_c=tsv_df,
    task_config={SparkConfig.conf: extend({"spark.default.parallelism": "2"})},
    result=parquet_df.output(name="filtered_data"),
)
def filter_data(
    data_a: DataFrame,
    data_b: DataFrame,
    data_c: DataFrame,
    extra_data: parameter[DataFrame],
    num_executors: int = 500,
) -> DataFrame:
    return data_a


@pipeline
def data_source_complicated_pipeline(root_path, extra_data):
    data = existing_data(
        root_path=root_path,
        config_ids_A=[1, 2],
        config_ids_B=[3, 4],
        config_ids_C=[5, 6],
    )
    return filter_data(data[0], data[1], data[2], extra_data=extra_data)



# © Copyright Databand.ai, an IBM Company 2022

from dbnd_examples_orchestration.orchestration.features.dynamic_tasks_with_pipeline import (
    say_hello_to_everybody,
)


class TestDynamicTasks(object):
    def test_run_say_hello_to_everybody(self):
        result = say_hello_to_everybody()
        assert (
            "Hey, user 2!",
            "Hey, some_user! and user 0 and user 1 and user 2",
        ) == result



# © Copyright Databand.ai, an IBM Company 2022

# This example requires pandas, numpy, sklearn, scipy
# Inspired by an MLFlow tutorial:
#   https://github.com/databricks/mlflow/blob/master/example/tutorial/train.py


import itertools
import logging

import numpy as np

from dbnd_examples_orchestration.data import data_repo
from pandas import DataFrame
from sklearn.linear_model import ElasticNet
from sklearn.metrics import mean_absolute_error, mean_squared_error, r2_score
from sklearn.model_selection import train_test_split

from dbnd import data, log_dataframe, log_metric, output, parameter
from dbnd_run.tasks import PipelineTask, PythonTask


logger = logging.getLogger(__name__)

test_data_csv = data_repo.wines


def _split(df):
    return df.head(int(len(df) / 2)), df.tail(int(len(df) / 2))


def calculate_metrics(actual, pred):
    rmse = np.sqrt(mean_squared_error(actual, pred))
    mae = mean_absolute_error(actual, pred)
    r2 = r2_score(actual, pred)
    return rmse, mae, r2


class PrepareData(PythonTask):
    """Split data into train, test and validation"""

    raw_data = parameter[DataFrame]

    training_set = output.csv
    test_set = output.csv
    validation_set = output.csv

    def run(self):
        self.log_dataframe("input", self.raw_data)

        train_df, test_df = train_test_split(self.raw_data)
        test_df, validation_df = train_test_split(test_df, test_size=0.5)

        self.log_dataframe("train_size", train_df)
        self.log_dataframe("test_size", test_df)
        self.log_dataframe("validation_df", validation_df)

        train_df.to_target(self.training_set, index=False)
        test_df.to_target(self.test_set, index=False)
        validation_df.to_target(self.validation_set, index=False)


class TrainModel(PythonTask):
    """Train wine prediction model"""

    alpha = parameter.value(0.5)
    l1_ratio = parameter.value(0.5)

    test_set = parameter[DataFrame]
    training_set = parameter[DataFrame]

    # fake parameter that represent quality
    quality = parameter.value(1.0)
    model = output.pickle.target

    def run(self):
        train = self.training_set
        test = self.test_set

        lr = ElasticNet(alpha=self.alpha, l1_ratio=self.l1_ratio)
        lr.fit(train.drop(["quality"], 1), train[["quality"]])
        prediction = lr.predict(test.drop(["quality"], 1))

        (rmse, mae, r2) = calculate_metrics(test[["quality"]], prediction)

        self.log_metric("train_size", len(train))
        self.log_metric("test_size", len(test))
        self.log_metric("rmse", rmse)
        self.log_metric("mae", rmse * self.quality)
        self.log_metric("r2", r2 * self.quality)

        self.model.write_pickle(lr)

        logging.info(
            "Elasticnet model (alpha=%f, l1_ratio=%f): rmse = %f, mae = %f, r2 = %f",
            self.alpha,
            self.l1_ratio,
            rmse,
            mae,
            r2,
        )


class ValidateModel(PythonTask):
    """Calculates metrics of wine prediction model"""

    model = parameter.data
    validation_dataset = parameter[DataFrame]
    model_metrics = output.csv.data

    def run(self):
        validation = self.validation_dataset
        self.log_metric("test_size", len(validation))

        actual_model = self.model.read_pickle()
        logger.info("%s", validation.shape)

        validation_x = validation.drop(["quality"], 1)
        validation_y = validation[["quality"]]

        prediction = actual_model.predict(validation_x)
        (rmse, mae, r2) = calculate_metrics(validation_y, prediction)

        log_dataframe("validation", validation)
        log_metric("rmse", rmse)
        log_metric("mae", rmse)
        log_metric("r2", r2)

        self.model_metrics.write("%s,%s,%s" % (rmse, mae, r2))


class PredictWineQuality(PipelineTask):
    """Entry point for wine quality prediction"""

    data = parameter(default=test_data_csv).data
    alpha = parameter.value(0.5)
    l1_ratio = parameter.value(0.5)

    model = output
    validation = output

    def band(self):
        datasets = PrepareData(raw_data=self.data)
        self.model = TrainModel(
            test_set=datasets.test_set,
            training_set=datasets.training_set,
            alpha=self.alpha,
            l1_ratio=self.l1_ratio,
        ).model

        self.validation = ValidateModel(
            model=self.model, validation_dataset=datasets.validation_set
        ).model_metrics


class PredictWineQualityParameterSearch(PipelineTask):
    data = data(default=test_data_csv).target
    alpha_step = parameter.value(0.3)
    l1_ratio_step = parameter.value(0.4)

    results = output

    def band(self):
        result = {}
        variants = list(
            itertools.product(
                np.arange(0, 1, self.alpha_step), np.arange(0, 1, self.l1_ratio_step)
            )
        )

        # variants = list(itertools.product([0.1, 0.5], [0.2, 0.3]))
        logger.info("All Variants: %s", variants)
        for alpha_value, l1_ratio in variants:
            predict = PredictWineQuality(
                data=self.data, alpha=alpha_value, l1_ratio=l1_ratio
            )

            exp_name = "%f_%f" % (alpha_value, l1_ratio)
            result[exp_name] = (predict.model, predict.validation)
        self.results = result



# © Copyright Databand.ai, an IBM Company 2022

import logging

import pytest

from dbnd_examples_orchestration.orchestration.features.partitioned_by_day import (
    ByDayExamplePipeline,
    FetchData,
    FetchIds,
    ProductionIdsAndData,
)

from dbnd import dbnd_config, get_databand_context
from dbnd_run.testing.helpers import assert_run_task


logger = logging.getLogger(__name__)


class TestPartitionsByDate(object):
    def test_prod_immutable_output_example(self):
        with dbnd_config(
            {FetchIds.task_enabled_in_prod: True, FetchData.task_enabled_in_prod: True}
        ):
            task = ProductionIdsAndData(
                task_env=get_databand_context().run_settings.env.clone(production=True)
            )
            assert_run_task(task)

    def test__by_day_simple_local(self):
        with dbnd_config(
            {
                ProductionIdsAndData.task_env: "local",
                FetchIds.task_enabled_in_prod: True,
                FetchData.task_enabled_in_prod: True,
            }
        ):
            assert_run_task(ByDayExamplePipeline(period="2d"))

    @pytest.mark.aws
    @pytest.mark.skip("aws permissions are wrong")
    def test__by_day_simple_aws(self):
        with dbnd_config(
            {FetchIds.task_enabled_in_prod: True, FetchData.task_enabled_in_prod: True}
        ):
            assert_run_task(ByDayExamplePipeline(period="2d"))



# © Copyright Databand.ai, an IBM Company 2022

import datetime
import logging

from pathlib import Path
from typing import List

import six

from dbnd import task
from targets import Target
from targets.types import PathStr


logger = logging.getLogger(__name__)


@task
def f_task_parameters(
    v_int: int,
    v_bool: bool,
    v_str: str,
    v_datetime: datetime.datetime,
    v_date: datetime.date,
    v_period: datetime.timedelta,
    v_list: List,
    v_list_obj: List[object],
    v_list_str: List[str],
    v_list_int: List[int],
    v_list_date: List[datetime.date],
) -> object:
    result = (
        v_int,
        v_datetime,
        v_str,
        v_date,
        v_period,
        v_list,
        v_list_obj,
        v_list_str,
        v_list_int,
        v_list_date,
    )
    logger.info("v_int: %s", v_int)
    logger.info("v_bool: %s", v_bool)
    logger.info("v_datetime: %s", v_datetime)
    logger.info("v_list: %s", v_list)
    logger.info("v_list_date: %s", v_list_date)

    return result


@task
def f_path_types(
    pathlib_path: Path, str_as_path: PathStr, target_path: Target
) -> object:
    assert isinstance(pathlib_path, Path)
    assert isinstance(str_as_path, six.string_types)
    assert isinstance(target_path, Target)

    return (pathlib_path, str_as_path, target_path)



# © Copyright Databand.ai, an IBM Company 2022

import time

from dbnd import pipeline, task


class TestDocOverridingConfigurationInCLI:
    def test_doc(self):
        #### DOC START
        @task
        def calculate_alpha(alpha=0.5):
            return alpha

        @pipeline
        def train_model_pipeline():
            calculate_alpha(0.1)

        #### DOC END

        train_model_pipeline.dbnd_run(task_version=time.time())



# © Copyright Databand.ai, an IBM Company 2022

import logging

from dbnd import get_databand_context, new_dbnd_context, task


logger = logging.getLogger(__name__)


def custom_logging_mod():
    base_config = get_databand_context().settings.log.get_dbnd_logging_config_base()

    base_config["handlers"]["my_file"] = {
        "class": "logging.FileHandler",
        "formatter": "formatter",
        "filename": "/tmp/my_custom_file",
        "encoding": "utf-8",
    }
    base_config["root"]["handlers"].append("my_file")

    return base_config


@task
def task_with_some_logging():
    logging.info("root logger")
    logger.info("custom logger")
    logger.warning("warning!")
    return "ok"


if __name__ == "__main__":
    mod_func = (
        "dbnd_examples_orchestration.feature_system.custom_logging.custom_logging_mod"
    )
    with new_dbnd_context(conf={"log": {"custom_dict_config": mod_func}}):
        task_with_some_logging.dbnd_run()



# © Copyright Databand.ai, an IBM Company 2022

from dbnd_examples_orchestration.orchestration.examples.report import (
    dump_db,
    partners_report,
)

from dbnd_run.testing.helpers import assert_run_task


class TestTutorialReportExamples(object):
    def test_tutorial_report_dump(self):
        task = assert_run_task(dump_db.task("aa,bb"))
        assert task

    def test_tutorial_partners_report(self):
        task = assert_run_task(partners_report.task([1, 2]))
        assert task
